/*
	PROforma, PROgs FOnt & Raster MAnager

	Routines which modify the CTM
	PFCTMSave
	PFCTMRestore
	PFCTMRestoreKeep
	PFCTMReset
	PFCTMSet
	user2dev
	reluser2dev

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 26 July 1993
	26.11.1996 PFCTMRestoreKeep added
*/

#include "gstate.h"
#include "system.h"
#include "list.h"

Error currCTMset(Gstate gstate)
{
	/* currCTM should also convert to device coordinates !! */
	gstate->currCTM=gstate->todefaultCTM;
	gstate->currCTM.xs=fixmul(gstate->currCTM.xs,gstate->todev.x);
	gstate->currCTM.ys=fixmul(gstate->currCTM.ys,gstate->todev.y);
	gstate->currCTM.xr=fixmul(gstate->currCTM.xr,gstate->todev.x);
	gstate->currCTM.yr=fixmul(gstate->currCTM.yr,gstate->todev.y);
	gstate->currCTM.xd=fixmul(gstate->currCTM.xd,gstate->todev.x);
	gstate->currCTM.yd=fixmul(gstate->currCTM.yd,gstate->todev.y);
	CacheFontRelease(&gstate->font);	/* release chached info - CTM changed */
	return ERR_OK;
}

Error PFCTMSave(Gstate gstate)
{
	CTMel *c;
	Error err;

	err=MEMAllocate(sizeof(CTMel),&c);
	if (err) return err;
	c->next=NULL;
	c->matrix=gstate->todefaultCTM;
	linkin(c,gstate->CTMlist);
	return err;
}

Error PFCTMRestore(Gstate gstate)
{
	CTMel *c;

	if (gstate->CTMlist==NULL) return ERR_ITNF;

	linkout(c,gstate->CTMlist);
	gstate->todefaultCTM=c->matrix;
	MEMRelease(c);

	return currCTMset(gstate);
}

Error PFCTMRestoreKeep(Gstate gstate)
{
	CTMel *c=gstate->CTMlist;

	if (c==NULL) return PFCTMSave(gstate);

	gstate->todefaultCTM=c->matrix;
	return currCTMset(gstate);
}

Error PFCTMReset(Gstate gstate)
{
	list_release(gstate->CTMlist);

	gstate->todefaultCTM.xs=gstate->todefaultCTM.ys=pt_one;
	gstate->todefaultCTM.xr=gstate->todefaultCTM.yr=
	gstate->todefaultCTM.xd=gstate->todefaultCTM.yd=0;

	return currCTMset(gstate);
}

Error PFCTMSet(Gstate gstate, pt xs, pt yr, pt xr, pt ys, pt xd, pt yd)
{
	CTM c;

	/* make a copy for the calculations */
	MEMCopy((long)sizeof(CTM), &gstate->todefaultCTM, &c);

	/* now calculate the new CTM */
	gstate->todefaultCTM.xs= fixmul(c.xs,xs)+fixmul(c.yr,xr);
	gstate->todefaultCTM.xr= fixmul(c.xr,xs)+fixmul(c.ys,xr);
	gstate->todefaultCTM.xd= fixmul(c.xd,xs)+fixmul(c.yd,xr)+xd;
	gstate->todefaultCTM.yr= fixmul(c.xs,yr)+fixmul(c.yr,ys);
	gstate->todefaultCTM.ys= fixmul(c.xr,yr)+fixmul(c.ys,ys);
	gstate->todefaultCTM.yd= fixmul(c.xd,yr)+fixmul(c.yd,ys)+yd;

	return currCTMset(gstate);
}

/* special aid routine, converts user coordinates to device coordinates */
void user2dev(Gstate gstate, coor *c)
{
	register coor *porg=&gstate->Origin;
	register CTM *m=&gstate->currCTM;
	register pt x=c->x, y=c->y;

	c->x = fixmul(x,m->xs) + m->xd - porg->x + fixmul(y,m->xr);
	c->y = fixmul(y,m->ys) + m->yd - porg->y + fixmul(x,m->yr);
}
/* special aid routine, converts user coordinates to device coordinates */
/* same as above, but for relative coordinates */
void reluser2dev(Gstate gstate, coor *c)
{
	register CTM *m=&gstate->currCTM;
	register pt x=c->x, y=c->y;

	c->x = fixmul(x,m->xs) + fixmul(y,m->xr);
	c->y = fixmul(y,m->ys) + fixmul(x,m->yr);
}
