/*
	PROforma, PROgs FOnt & Raster MAnager

	PROforma runtime configuration handling
	PFConfig	configure PROforma - pass configuration line

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 21 September 1995
	12.07.1996 driver for config is now either the last linked driver
			   (as before), or if the default driver is set, then that
			   driver will receive the config lines
	02.08.1996 The set path options can now start with a '+' in which case the
				 path is added to the previous.
	06.08.1996 added 'c' to configure no of colour caches for each gstate
	04.12.1997 test for version now possible (using 'v')
*/

#include <str.h>
#include <mem.h>
#include <job.h>
#include <module.h>
#include <atomic.h>
#include <ctype.h>

#include "gstate.h"
#include "system.h"

#define PF_DRIVER "PROforma external driver"
#define MAX_PATH 256
#define FONTMAP_BLOCK_FACTOR 50

extern Job PROforma;				/* PROforma Job identifier */
extern BITMdriver dummybits;		/* dummy bitmap driver */

static Chunks fontmapchunks=NUL;	/* Chunks for fontmap */
fontline *fontmap=NULL; 			/* map of existing fonts */

char fontpath[MAX_PATH]=""; 		/* search path for fonts */
static char drivpath[MAX_PATH]="";	/* search path for drivers */

int memory=-65536;					/* memory usage when opening drivers */
int ColourCacheCount=8; 			/* no of entries in each colour cache */

DRIVdriver *printers=NULL;			/* currently supported printer drivers */
DRIVdriver *screen=NULL;			/* current screen driver */
DRIVdriver *defdev=NULL;			/* current default printer driver */
PICTdriver *pictures=NULL;			/* currently supported picture drivers */
BITMdriver *bitmaps=&dummybits; 	/* currently supported bitmap drivers */

static DRIVdriver *config=NULL; 	/* driver for config, last or default */

static Error FontLink(char *p)
{
	Error err;
	fontline *fl, *list;
	char *q;

	err=MEMUserAllocate(fontmapchunks,&fl);
	if (err) return err;

	q=fl->FullName;
	while (*p!=';' && *p!='\0' && *p!='\n' && *p!='\r') *q++=*p++;
	*q='\0';
	p++;
	q=fl->filename;
	while (*p!=';' && *p!='\0' && *p!='\n' && *p!='\r') *q++=*p++;
	*q='\0';

	/* link it in */

	fl->next=NULL;
	if (fontmap==NULL)
		fontmap=fl;
	else
	{
		for (list=fontmap ; list->next ; list=list->next);
		list->next=fl;
	}
	return err;
}

static Error PictLink(PICTdriver *pict)
{
	PICTdriver *p=pict;
	while (p->next) p=p->next;
	p->next=pictures;
	pictures=pict;
	return ERR_OK;
}

static Error BitmLink(BITMdriver *bitm)
{
	BITMdriver *b=bitm;
	while (b->next) b=b->next;
	b->next=bitmaps;
	bitmaps=bitm;
	return ERR_OK;
}

static Error PrtLink(DRIVdriver *prt)
{
	DRIVdriver *l;
	if (l=printers)
	{
		while (l->next) l=l->next;
		l->next=prt;
	}
	else
	{
		printers=prt;
		defdev=prt;
	}
	return ERR_OK;
}

static Error ScrLink(DRIVdriver *scr)
{
	screen=scr;
	return ERR_OK;
}

Error PFConfig(char *line)
{
	Error err=ERR_OK;

	/* fontmapchunks have to be allocated by PROforma job itself !! */
	/* also the cache has to be initialised to be in a safe state */
	if (!fontmapchunks)
	{
		err=MEMUserInit(sizeof(fontline),FONTMAP_BLOCK_FACTOR,&fontmapchunks);
		if (err) return err;
		err=CacheInit();
		if (err) return err;
	}

	/* just skip empty and comment lines */
	if (line[0]=='\0' || line[0]==';' || line[0]=='%') return err;

	if (line[1]==' ')
	{
		/* command for configurator */
		switch(line[0])
		{
		case 'S':
			/* default search path for fonts */
			if (*(line+2)=='+')
			{
				STRAppendSafe(";",fontpath,MAX_PATH);
				STRAppendSafe(line+3,fontpath,MAX_PATH);
			} else
				STRCopySafe(line+2,fontpath,MAX_PATH);
			break;
		case 's':
			/* default search path for drivers */
			if (*(line+2)=='+')
			{
				STRAppendSafe(";",drivpath,MAX_PATH);
				STRAppendSafe(line+3,drivpath,MAX_PATH);
			} else
				STRCopySafe(line+2,drivpath,MAX_PATH);
			break;
		case 'D':
			{
				ANYdriver *driv;

				err=MODULELoadFor(line+2,drivpath,PROforma,PF_DRIVER,&driv);
				if (err) break;

				switch(driv->identifier)
				{
				case PF_SCREENDRIVER:
					err=((DRIVdriver *)driv)->InitDriver();
					if (err) return err;
					ATOMICCall(ScrLink,1,driv);
					break;
				case PF_PRINTERDRIVER:
					config=(DRIVdriver *)driv;
					err=((DRIVdriver *)driv)->InitDriver();
					if (err) return err;
					ATOMICCall(PrtLink,1,driv);
					break;
				case PF_BITMAPDRIVER:
					ATOMICCall(BitmLink,1,driv);
					break;
				case PF_PICTUREDRIVER:
					ATOMICCall(PictLink,1,driv);
					break;
				default:
					err=ERR_ISYN;
					break;
				}
				break;
			}
			break;
		case 'C':
			/* set size of fontcache */
			{
				int length, count;
				length=count=0;
				STRFormatGet(line+2,NULL,"%d %d",&length,&count);
				err=CacheAllocate(length,count);
				if (err) break;
			}
			break;
		case 'c':
			/* set size of colour cache */
			STRFormatGet(line+2,NULL,"%d",&ColourCacheCount);
			if (ColourCacheCount>256) ColourCacheCount=256;
			if (ColourCacheCount<2) ColourCacheCount=1;
			break;
		case 'M':
			/* set memory configuration */
			STRFormatGet(line+2,NULL,"%d",&memory);
			break;
		case 'P':
			err=ATOMICCall(FontLink,1,line+2);
			break;
		case 'R':
			/* load a resident font */
			/* the parameter is the font file name */
			err=FontResidentLoad(line+2);
			break;
		case 'd':
			/* select the default printer driver */
			{
				char *start=line+2;
				DRIVdriver *driver;
				int id, driverID;

				while (isspace(*start)) start++;
				if (*start=='-')
				{
					/* driver id has been passed */
					STRFormatGet(start,NULL,"%d",&driverID);
					for (driver=printers, id=-2 ;
						 driver && id!=driverID ;
						 driver=driver->next, id--);
				}
				else
				{
					/* driver name has been passed */
					for (driver=printers ;
						 driver && !STRSameCD(driver->name, start) ;
						 driver=driver->next);
				}
				if (driver) defdev=config=driver;
			}
			break;
		case 'v':
			/* check for a minimum version */
			/* gives ERR_ISYN if version is older then requested */
			/* otherwise returns ERR_OK */
			{
				char name[JOB_MAXNAME];
				char *version, *requested=line+2;
				MEMClear(JOB_MAXNAME,name); /* to cater for old syslibs */
				JOBInfo(PROforma, name, NULL, NULL);
				version=name+strlen(name)+2;
				while (*requested && *requested==*version)
					requested++, version++;
				if (*requested>*version) return ERR_ISYN;
				return ERR_OK;
			}
			break;
		default:
			err=ERR_IPAR;
		}
	}
	else
	{
		/* full blown define - only passed to the last added type */
		if (config)
		{
			char *first, *second;

			/* now this line has to be split in keyword/data */
			second=first=line;
			while (*second && !isspace(*second)) second++;
			if (isspace(*second))
				*second++='\0';
			while (isspace(*second)) second++;

			err=config->Define(first,second);
		}
	}
	return err;
}
