/*
	PROforma, PROGS Font & Raster Manager

	PROforma thing handling routines

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 20 September 1995
	18.09.1996 thing now always only used once by each job
*/

#include "thing.h"
#include "job.h"

#include "PROforma.h"

extern char InitIsDone; 			/* to see if linking is safe to do */
extern Job PROforma;				/* job id for myself */

Error PFConfig(char *line); 		/* routine for runtime configuration */

typedef struct {
	char identifier[8]; 			/* e.g. "<<QDLL>>" */
	void *usercode; 				/* base address for usercode */
	int extension;					/* name of extension in thing */
	short namelen;					/* length of thingname == 12 */
	char thingname[12]; 			/* "PROforma DLL" */
	int version;					/* " v01" == PF_DLL_VERSION */
	short links[1]; 				/* list of functions to link (VECT only) */
	} DLLEntry;

#define PF_THING_CORE		0x434f5245	/* "CORE" */
#define PF_THING_VECT		0x56454354	/* "VECT" */
#define PF_THING_VERSION	0x312e3130	/* "1.10" */
#define PF_DLL_VERSION		0x20763031	/* " v01" */

#define DLL_TYPE_MASK	(short)0xc000
#define DLL_ID_MASK 	(short)0x3fff
#define DLL_TYPE_JMPL	(short)0x4000
#define DLL_TYPE_DATA	(short)0x8000
#define DLL_JMPL		(short)0x4ef9

static Error DLLink(DLLEntry *link)
{
	Error err;
	Job me, user;
	int counter, count;

	/* check version */
	if (link->version!=PF_DLL_VERSION)
		return ERR_NIMP;

	JOBMe(&me);
retry:
	counter=0; count=0;
	while (!(err=THINGUserNext(PF_THING_NAME,&counter,&user)))
	{
		if (user==me)
		{
			count++;
			if (count>1) THINGFree(PF_THING_NAME);
		}
		if (counter==0) break;
	}
	if (err==ERR_IJOB) goto retry;

	if (me!=PROforma)				/* I can always link back to myself */
	{
		/* wait until initialisation has finished */
		while (!InitIsDone)
			JOBSuspend(MYSELF,1,NULL);
	}

	/* now link the proper area */
	switch (link->extension)
	{
	case PF_THING_CORE:
		{
			extern char core_start[2], core_end[2];
			char *dummy=core_end;
			MEMCopy(dummy-core_start,core_start,link->usercode);
		}
		break;
	case PF_THING_VECT:
		{
			extern void *DLLARRAYpf[2];
			extern short COUNTpf;
			void *dest=(void *)((char *)link->usercode+2);
			short *what, id;

			for (what=link->links ; *what ; what++)
			{
				id=(*what&DLL_ID_MASK)-1;
				if (id>=COUNTpf)
					return ERR_NIMP;
				switch(*what&DLL_TYPE_MASK)
				{
				case DLL_TYPE_JMPL:
					*((short *)dest)++=DLL_JMPL;
					*((void **)dest)++=DLLARRAYpf[id];
					break;
				case DLL_TYPE_DATA:
					*((void **)dest)++=*((void **)DLLARRAYpf[id]);
					break;
				default:
					return ERR_NIMP;
				}
			}
		}
		break;
	default:
		return ERR_IPAR;
	}
	return ERR_OK;
}

static ThingExtension thg_list[]={ { PF_THING_VECT, DLLink		},
								   { PF_THING_CORE, DLLink      },
								   { PF_THING_CNFG, PFConfig    },
								   { THING_LISTSENTINEL         } };

static ThingDefinition thg_defn={PF_THING_NAME, PF_THING_VERSION, FALSE,
								 NULL, NULL, NULL, NULL, 0, thg_list};

Error PFThingInit()
{
	Error err;
	Thing thing;

	/* remove old thing, unless still in use */
	err=THINGRemove(PF_THING_NAME);
	if (err && err!=ERR_ITNF) return err;

	/* link in new one */
	thg_defn.owner=PROforma;
	err=THINGCreate(&thg_defn);
	if (err) return err;

	/* make sure the PROforma job uses the PROforma thing */
	return THINGUse(PF_THING_NAME,0,&thing,NULL);
}
