/*
	PROforma, PROGS Font & Raster Manager

	Epson 9 pin printer driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 24 January 1996, based on DeskJet driver
*/

#include "mem.h"
#include "driv/escp.h"

#define PINS 24

Bbox PrintableArea={
	double2pt(577.8), short2pt(780),/* size */
	double2pt(8.6), short2pt(24)	/* origin */
};

char PrintDevice[IO_MAXFULLNAME]="pard";

Error Define(char *name, char *value)
{
	double x,y;

	if (STRSameCD(name,"DEFAULT-DEVICE"))
	{
		STRCopySafe(value,PrintDevice,IO_MAXFULLNAME);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-SIZE"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xsiz=double2pt(x);
		PrintableArea.ysiz=double2pt(y);
		return ERR_OK;
	}
	if (STRSameCD(name,"PRINTABLE-AREA-ORIGIN"))
	{
		STRFormatGet(value,NULL,"%f %f",&x,&y);
		PrintableArea.xorg=double2pt(x);
		PrintableArea.yorg=double2pt(y);
		return ERR_OK;
	}
	return ERR_IPAR;
}

Error InitInit(Gstate gstate, char *device, PBbox *area)
{
	Error err;
	DriverInfo *di;
	DRIVdriver *prt=gstate->printer;

	err=MEMAllocate(sizeof(DriverInfo),&di);
	if (err) return err;

	di->pins=PINS;					/* no of pins which are used */
	di->options=0;					/* options (0=none) */
	if (gstate->printer->ydpi==360) di->options+=HIGH;

	if (device && *device)
		err=IOOpen(device,OPEN_OVER,&di->channel);
	else
		err=IOOpen(PrintDevice,OPEN_OVER,&di->channel);
	if (err) { MEMRelease(di); return err; };

	gstate->PrinterExtra=di;
	gstate->todev.x = pt_one * prt->xdpi / 72;
	gstate->todev.y = pt_one * prt->ydpi / 72;
	gstate->ppp.x = pt_one * 72 / prt->xdpi;
	gstate->ppp.y = pt_one * 72 / prt->ydpi;

	area->xsiz=pt2rshort(fixmul(PrintableArea.xsiz,gstate->todev.x));
	area->ysiz=pt2rshort(fixmul(PrintableArea.ysiz,gstate->todev.y));
	area->xorg=pt2rshort(fixmul(PrintableArea.xorg,gstate->todev.x));
	area->yorg=pt2rshort(fixmul(PrintableArea.yorg,gstate->todev.y));

	return err;
}

Error InitDone(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;
	Size linc;

	gstate->bitmap->BufferSize(gstate->pagebbox.xsiz,
					di->options&HIGH?PINS*2:PINS,&linc);

	/* complete the initialisation of the DriverInfo */
	di->buffer=(char *)gstate->raster->base;
	gstate->raster->base+=linc;

	di->xskip=pt2rshort(gstate->pagebbox.xporg-fixmul(PrintableArea.xorg,
							   gstate->todev.x));
	di->yskip=pt2rshort(gstate->pagebbox.yporg-fixmul(PrintableArea.yorg,
							   gstate->todev.y));

	return ERR_OK;
}

Error BufferSize(Gstate gstate, int xsiz, int ysiz, Size *size)
{
	DriverInfo *di=gstate->PrinterExtra;
	return gstate->bitmap->BufferSize(xsiz,ysiz+(di->options&HIGH?PINS*2:PINS),size);
}

Error CleanUp(Gstate gstate)
{
	DriverInfo *di=gstate->PrinterExtra;

	if (di)
	{
		IOClose(di->channel);
		MEMRelease(gstate->PrinterExtra);
	}
	return ERR_OK;
}

Error ColourNearest(Gstate gstate, pt *colour, int *result)
{
	*result=0;						/* assume white */
	if (*colour>pt_hundred/2)
		*result=1;					/* make it black */
	return ERR_OK;
}

Error ColourDelta(Gstate gstate, pt *colour, pt *delta)
{
	*delta=*colour;
	if (*colour>pt_hundred/2)
		*delta-=pt_hundred;
	return ERR_OK;
}

static Error nimp() 	{ return ERR_NIMP; }
static Error nothing()	{ return ERR_OK; }

Error InitDriver();

Error SetCopies(Gstate, int count);
Error SetDisplayMode(Gstate, int mode);
Error WindowMove(Gstate, int dx, int dy);
Error BufferUpdate(Gstate);
Error InitPage(Gstate);
Error ShowPage(Gstate);
Error StopPage(Gstate);
Error ColourAdjust(Gstate, pt *);
Error GetFullPageBbox(Gstate, Bbox *box);
Error DefaultDevice(Gstate, char *device);

#define InitDriver		nothing
#define WindowMove		nothing
#define ColourAdjust	nothing
#define SetDisplayMode	nimp
#define GetFullPageBbox nimp
#define DefaultDevice	nimp
#define SetCopies		nimp
#define BufferUpdate	(Error (*)(Gstate))nimp
#define BufferSave		(Error (*)(Gstate))nimp
#define BufferRestore	(Error (*)(Gstate))nimp
#define Handle			(Error (*)(int,...))nimp

DRIVdriver dpi60 = {
	NULL, PF_PRINTERDRIVER,
	"Epson Compatible 24 pin, 60dpi",	/* name */
	60,60,						/* xdpi, ydpi */
	{                           /* mono bitmap */
		1, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_NO_PLANES,		/* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_GRAYSHADE,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};

DRIVdriver dpi180 = {
	&dpi60, PF_PRINTERDRIVER,
	"Epson Compatible 24 pin, 180dpi",	/* name */
	180,180,					/* xdpi, ydpi */
	{                           /* mono bitmap */
		1, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_NO_PLANES,		/* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_GRAYSHADE,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};

DRIVdriver dpi360 = {
	&dpi180, PF_PRINTERDRIVER,
	"Epson Compatible 24 pin, 360x180dpi",	/* name */
	360,180,					/* xdpi, ydpi */
	{                           /* mono bitmap */
		1, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_NO_PLANES,		/* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_GRAYSHADE,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};

DRIVdriver init = {
	&dpi360, PF_PRINTERDRIVER,
	"Epson Compatible 24 pin, 360dpi",	/* name */
	360,360,					/* xdpi, ydpi */
	{                           /* mono bitmap */
		1, 1,					/* 1 plane, 1 pixel deep */
		BITMAP_NO_BYTEGROUPING, /* no byte grouping */
		BITMAP_NO_PLANES,		/* no planes */
	},

	InitDriver,
	Define,

	InitInit,
	BufferSize,
	InitDone,
	CleanUp,

	InitPage,
	ShowPage,
	StopPage,

	SetCopies,
	SetDisplayMode,
	WindowMove,

	BufferUpdate,
	BufferSave,
	BufferRestore,

	PF_COLOURSPACE_GRAYSHADE,
	{
		ColourAdjust,
		ColourNearest,
		ColourDelta
	},

	GetFullPageBbox,
	DefaultDevice,

	Handle
};
