/*
	PROforma, PROgs FOnt & Raster MAnager

	LINEdesign v2.0 picture driver

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 9 November 1996
	27.11.1996 working
	03.08.1997 made safer (but does need extra memory now) on display of
			   paragraph text and bitmaps
*/

#include <str.h>
#include <math.h>
#include "PFmodule.h"

typedef struct {
	pt xorg, yorg;
	pt xend, yend;
} Bounds;

typedef struct {
	char id[15];					/* "LINEdesign v2.0" - no ending zero !! */
	char bbox[16];
	char page[16];
	char BitmapCount[4];
	char bitmaps[1];				/* &bitmaps is start of bitmap table */
} Page;

#define 	f_path			1
#define 	f_path_start	2
#define 	f_path_line 	3
#define 	f_path_curve	4
#define 	f_path_end		5
#define 	f_circle		6
#define 	f_square		7
#define 	f_rectangle 	8
#define 	f_group 		9
#define 	f_header		10
#define 	f_paragraph 	11
#define 	f_bitmap_head	12
#define 	f_bitmap_image	13
#define 	f_bitmap_obj	14

#define UseRead 		typedef struct { char c1,c2,c3,c4; } char4; \
						union { char4 c; short s; long l; } mix;
#define readlong(var)	mix.c.c1=*run++; mix.c.c2=*run++; mix.c.c3=*run++; \
						mix.c.c4=*run++; var=mix.l;
#define readshort(var)	mix.c.c1=*run++; mix.c.c2=*run++; var=mix.s;
#define HandleTrafo run=_HandleTrafo(gstate,run,&trafo)

static char *_HandleTrafo(Gstate gstate, char *run, Bbox *trafo)
{
	pt x1,x2,x3,y1,y2,y3;
	UseRead;
	
	readlong(x1); readlong(y1);
	readlong(x2); readlong(y2);
	readlong(x3); readlong(y3);
	PFCTMRestoreKeep(gstate);
	PFCTMSet(gstate,x1,y1,x2,y2,x3,y3);
	PFCTMMove(gstate,trafo->xorg,trafo->yorg);
	PFCTMSet(gstate,trafo->xsiz,0,0,trafo->ysiz,0,0);
	return run;
}

static Error Handle(int command,...)
{
	return ERR_NIMP;
}

Error Recognize(Gstate gstate, char *base)
{
	FileInfo *fi=(FileInfo *)base;
	Page *pict=(Page *)(base+sizeof(FileInfo));

	if (fi->type==FILETYPE_NORMAL &&
		fi->length>sizeof(Page) &&
		MEMSame(15,pict->id,"LINEdesign v2.0"))
		return ERR_OK;
	else
		return ERR_INAM;
}

Error AspectRatio(Gstate gstate, char *base, int *xratio, int *yratio)
{
	Page *pict=(Page *)(base+sizeof(FileInfo));
	pt xorg,xend,yorg,yend;
	char *run=pict->bbox;
	UseRead;

	LinkPROforma();
	readlong(xorg);
	readlong(yorg);
	readlong(xend);
	readlong(yend);
	*xratio=xend-xorg;
	*yratio=yend-yorg;
	return ERR_OK;
}

Error Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
					 Error (*ColourSelect)(Gstate, int))
{
	FileInfo *fi=(FileInfo *)base;
	Page *pict=(Page *)(base+sizeof(FileInfo));
	char *bitmaps=pict->bitmaps;
	char *run=pict->bbox, *end=(char *)pict+fi->length;
	int i,len,id;
	char *strokestart;
	char closed, isfilled;
	short outlcol, fillcol;
	pt x1,x2,x3,y1,y2,y3,x,y;
	pt xorg,xend,yorg,yend;
	Bbox trafo;
	Error err;
	UseRead;

	readlong(trafo.xorg);
	readlong(trafo.yorg);
	readlong(trafo.xsiz);			/* actually xend */
	readlong(trafo.ysiz);			/* actually yend */

	/* clear background behind picture */
	PFColourGray(gstate,0);
	PFPathFilled(gstate);
	PFMoveTo(gstate,0,0);
	PFRectangle(gstate,xsiz,ysiz,0);
	PFPathDraw(gstate);

	/* scale the picture to the right size */
	trafo.xsiz=double2pt((double)xsiz/(trafo.xsiz-trafo.xorg));
	trafo.ysiz=double2pt((double)ysiz/(trafo.ysiz-trafo.yorg));
	trafo.xorg=-trafo.xorg;
	trafo.yorg=-trafo.yorg;

	/* skip all the bitmaps */
	run=pict->BitmapCount;
	readlong(i);
	for ( ; i>0 ; i--)
	{
		readlong(len);
		run+=len;
	}

	PFCTMSave(gstate);
	while (run<end)
	{
		switch (*run++)
		{
		case f_path:
			run+=16;				/* skip bbox */
			readlong(x);
			PFLineWidth(gstate,fixmul(x,trafo.xsiz));
			isfilled=*run++;
			strokestart=run;
strokepath:
			if (isfilled)
			{
				run+=2;
				readshort(x);
				PFPathEOFilled(gstate);
			}
			else
			{
				readshort(x);
				run+=2;
				PFPathStroked(gstate);
			}
			PFColourGray(gstate,short2pt(x));
			HandleTrafo;
			closed=*run++;
			break;
		case f_path_start:
			readlong(x);
			readlong(y);
			PFMoveTo(gstate,x,y);
			readlong(x3);			/* next controlpoint in case of curve */
			readlong(y3);
			x3+=x;
			y3+=y;
			run++;					/* skip nodetype */
			break;
		case f_path_line:
			run+=8; 				/* skip previous controlpoint */
			readlong(x);
			readlong(y);
			PFLineTo(gstate,x,y);
			readlong(x3);			/* next controlpoint in case of curve */
			readlong(y3);
			x3+=x;
			y3+=y;
			run++;					/* skip nodetype */
			break;
		case f_path_curve:
			readlong(x1);
			readlong(y1);
			readlong(x);
			readlong(y);
			PFCurveTo(gstate,x3,y3,x+x1,y+y1,x,y);
			readlong(x3);			/* next controlpoint in case of curve */
			readlong(y3);
			x3+=x;
			y3+=y;
			run++;					/* skip nodetype */
			break;
		case f_path_end:
			if (closed) PFPathClose(gstate);
			PFPathDraw(gstate);
			if (isfilled)
			{
				isfilled=FALSE;
				run=strokestart;
				goto strokepath;
			}
			break;
		case f_circle:
			run+=16;				/* skip bbox */
			readlong(x);
			PFLineWidth(gstate,fixmul(x,trafo.xsiz));
			isfilled=*run++;
			readshort(outlcol);
			readshort(fillcol);
			HandleTrafo;
			readlong(x); readlong(y);
			readlong(len);
			readlong(x1); readlong(y1);
			readlong(x2); readlong(y2);
			run++;
			closed=*run++;
			if (isfilled)
			{
				PFPathEOFilled(gstate);
				PFColourGray(gstate,short2pt(fillcol));
				PFMoveTo(gstate,x,y);
				PFPie(gstate,len,x1,y1,x2,y2,closed);
				PFPathDraw(gstate);
			}
			PFPathStroked(gstate);
			PFColourGray(gstate,short2pt(outlcol));
			PFMoveTo(gstate,x,y);
			PFPie(gstate,len,x1,y1,x2,y2,closed);
			PFPathDraw(gstate);
			break;
		case f_square:
			run+=16;				/* skip bbox */
			readlong(x);
			PFLineWidth(gstate,fixmul(x,trafo.xsiz));
			isfilled=*run++;
			readshort(outlcol);
			readshort(fillcol);
			HandleTrafo;
			readlong(x); readlong(y);
			readlong(x1); readlong(y1);
			run++;
			if (isfilled)
			{
				PFPathEOFilled(gstate);
				PFColourGray(gstate,short2pt(fillcol));
				PFMoveTo(gstate,x,y);
				PFRectangle(gstate,x1,x1,y1);
				PFPathDraw(gstate);
			}
			PFPathStroked(gstate);
			PFColourGray(gstate,short2pt(outlcol));
			PFMoveTo(gstate,x,y);
			PFRectangle(gstate,x1,x1,y1);
			PFPathDraw(gstate);
			break;
		case f_rectangle:
			run+=16;				/* skip bbox */
			readlong(x);
			PFLineWidth(gstate,fixmul(x,trafo.xsiz));
			isfilled=*run++;
			readshort(outlcol);
			readshort(fillcol);
			HandleTrafo;
			readlong(x); readlong(y);
			readlong(x1); readlong(y1);
			readlong(len);
			run++;
			if (isfilled)
			{
				PFPathEOFilled(gstate);
				PFColourGray(gstate,short2pt(fillcol));
				PFMoveTo(gstate,x,y);
				PFRectangle(gstate,x1,y1,len);
				PFPathDraw(gstate);
			}
			PFPathStroked(gstate);
			PFColourGray(gstate,short2pt(outlcol));
			PFMoveTo(gstate,x,y);
			PFRectangle(gstate,x1,y1,len);
			PFPathDraw(gstate);
			break;
		case f_header:
			run+=16;				/* skip bbox */
			HandleTrafo;
			readshort(x);
			PFColourGray(gstate,short2pt(x));
			readlong(x);
			readlong(y);
			PFMoveTo(gstate,x,y);
			readlong(x);
			PFFontScale(gstate,x);
			PFFontLoad(gstate,run);
			while (*run++);
			readlong(len);
			{
				short w[2];
				static const pt nul=0;
				char *l1=run-2;
				char *l2=run+len*2-4;
				char *l3=run+len*6-4;
				char *hend=run+len*10;
				w[1]=0;
				for (i=len ; i>0 ; i--)
				{
					run=l1+=2;
					readshort(w[0]);
					run=l2+=4;
					readlong(x3);
					run=l3+=4;
					readlong(y3);
					PFMoveR(gstate,x3,y3);
					PFShowXY(gstate,w,&nul,&nul);
				}
				run=hend;
			}
			break;
		case f_paragraph:
			run+=16;				/* skip bbox */
			HandleTrafo;
			readshort(x);
			PFColourGray(gstate,short2pt(x));
			{
				pt xorg, yorg, linewidth;
				int justify, kern;
				pt spacing, leading, indent;
				short *runner, *crunner, *endrunner;
				char *start;
				int lines, newpara, lastline;
				readlong(xorg);
				readlong(yorg);
				readlong(linewidth);
				linewidth-=xorg;
				readlong(lines);
				readlong(justify);
				readlong(kern);
				readlong(spacing);
				readlong(leading);
				readlong(indent);
				readlong(x);
				PFFontScale(gstate,x);
				PFFontLoad(gstate,run);
				while (*run++);
				readlong(len);
				err=MEMAllocate(len*2,&start);
				if (!err)
				{
					MEMCopy(len*2,run,start);
					runner=(short *)start;
					endrunner=runner+len;
					newpara=TRUE;
					PFExtraEOS(gstate,10);
					for ( ;
						 runner<endrunner && yorg+leading<lines ;
						 yorg+=leading)
					{
						/* move to next line (to check whether this is last) */
						crunner=runner;
						while (*crunner && *crunner!='\n') crunner++;
						lastline=(*++crunner=='\n' ||*crunner=='\0');

						switch(justify)
						{
						case 0: /* left */
left:						PFMoveTo(gstate,xorg+(newpara?indent:0),yorg);
							if (kern)
								PFShowKern(gstate,runner);
							else
								PFShow(gstate,runner);
							break;
						case 1: /* right */
						case 2: /* center */
							if (kern)
								PFWidthKern(gstate,runner,&x3);
							else
								PFWidth(gstate,runner,&x3);
							PFMoveTo(gstate,xorg+(linewidth-x3)/justify,yorg);
							if (kern)
								PFShowKern(gstate,runner);
							else
								PFShow(gstate,runner);
							break;
						case 3: /* full */
							PFMoveTo(gstate,xorg+(newpara?indent:0),yorg);
							if (lastline) goto left;
							if (kern)
								PFShowKernJust(gstate,runner,linewidth-(newpara?indent:0));
							else
								PFShowJust(gstate,runner,linewidth-(newpara?indent:0));
							break;
						}
						newpara=lastline;
						runner=crunner;
						
						if (newpara) yorg+=spacing, runner++;
					}
					MEMRelease(start);
				}
				run+=len*2; 	/* make run skip text */
				readlong(len);
				if (len) while(*run++);
			}
			break;
		case f_bitmap_obj:
			run+=16;				/* skip bbox */
			HandleTrafo;
			readlong(x);			/* position of bitmap */
			readlong(y);
			PFMoveTo(gstate,x,y);
			readlong(x3);			/* size for bitmap */
			readlong(y3);
			readlong(id);			/* screentype */
			while (*run++); 		/* skip filename */
			readlong(len);			/* ColourCount */
			run+=len*4+4;			/* skip colours & bitmaplen */
			readlong(i);			/* index in bitmaptable */
			{
				char *keeprun=run, *start;

				run=pict->bitmaps;
				for( ; i>0 ; i--)
				{
					readlong(len);
					run+=len;
				}
				readlong(len);

				err=MEMAllocate(len,&start);
				if (!err)
				{
					MEMCopy(len,run,start);
					PFPictureDisplay(gstate,id,start-sizeof(FileInfo),x3,y3);
					MEMRelease(start);
				}
				run=keeprun;		/* restore position in picture */
			}
			break;
		case f_group:
			run+=16+4;				/* just skip header and draw on */
			break;
		}
	}
	PFCTMRestore(gstate);
	return ERR_OK;
}

Error ColourInfo(Gstate gstate, char *base, int *count, int *colourspace)
{
	*count=0;
	return ERR_OK;
}

Error ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	return ERR_OK;
}

PICTdriver init = {
	NULL, PF_PICTUREDRIVER,
	"LINEdesign v2 page",

	/* can we recognize the picture */
	Recognize,
	AspectRatio,

	/* display the picture */
	Display,

	/* info on colours used in the picture */
	ColourInfo,
	ColourGet,

	Handle
};
