/* PROforma picture driver, to display GIF files */
/* based on QGIF */
/* this driver is flawed, as it doesn't allow you to replace the colours */

/* As this is a GIF driver, it uses the LZW algorithm to decompress */
/* the picture. However, according to the FSF (Free Software Foundation, */
/* best known as the authors of Ghostscript and makers of a free OS), the */
/* LZW patents owned by Unisys and IBM do not cover decompression. Thus */
/* you are allowed to use this software. */

/* The driver is only supposed to work on GIF87a files, but does also */
/* recognize GIF89a files as they are usually also displayable. */
/* The driver can handle interleaved pictures nicely. */
/* there are some gif files which can't be displayed with this driver */
/* this is probably caused by some unsupported block type - it was a GIF89a */

/* When sysmon starts whaling at you during display of a picture, this */
/* usually means you ran out of stack space. */

/* To increase the speed at which GIF files are displayed, it is interesting */
/* to increase the size of the colour stack to 256 colours. */

#include "PFmodule.h"

Error PFPictureElement(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg);

typedef struct codestruct {
	struct codestruct *prefix;
	unsigned char first,suffix;
} codetype;

#define BYTE2PERCENT(x) (pt) ((int)( ((x)>100 ? (x)+1 : (x)) ) * 25600 )

typedef struct {
	unsigned int width, height; 	/* width and height of picture */
	int interleaved;				/* is picture interleaved ?? */
	pt xsiz, ysiz, totsiz;			/* size of a pixel */
	int line, row;					/* line and row for next pixel */
	pt xpos, ypos;					/* position for next pixel */
	Gstate gstate;					/* gstate for display */
	codetype *codetable;			/* LZW compression code data */
	int datasize,codesize,codemask; /* Decoder working variables */
	int avail, oldcode; 			/* static var for process */
	int clear, eoi; 				/* Special code values */
	ColourRGB colour;				/* for outcode routine */
	int drawpixel;					/* for outcode - is pixel visible */
	unsigned char colourmap[256][3];/* Intensity of colors */
} picturedata;


static Error Handle(int command,...)
{
	LinkPROforma();
	return ERR_NIMP;
}

static Error Recognize(Gstate gstate, char *base)
{
	if (MEMSame(6,base+sizeof(FileInfo),"GIF87a") ||
		MEMSame(6,base+sizeof(FileInfo),"GIF89a"))
		return ERR_OK;
	else
		return ERR_INAM;
}

static Error AspectRatio(Gstate gstate, char *base, int *xratio, int *yratio)
{
	unsigned char *buf=(unsigned char *)base+sizeof(FileInfo)+6;

	/* value in case we can't find xsiz/ysiz */
	*xratio = buf[0] + (buf[1] << 8);
	*yratio = buf[2] + (buf[3] << 8);
	return ERR_OK;
}

static Error ColourInfo(Gstate gstate, char *base, int *count, int *space)
{
	*count=0;
	*space=PF_COLOURSPACE_RGB;
	return ERR_OK;
}

static Error ColourGet(Gstate gstate, char *base, int which, void *colour)
{
	return ERR_IPAR;
}

/* Output the bytes associated with a code to the raster array */

void outcode(picturedata *pd, codetype *p)
{
	if (p->prefix) outcode(pd,p->prefix);
	pd->colour.red	=BYTE2PERCENT(pd->colourmap[p->suffix][0]);
	pd->colour.green=BYTE2PERCENT(pd->colourmap[p->suffix][1]);
	pd->colour.blue =BYTE2PERCENT(pd->colourmap[p->suffix][2]);
	if (pd->xpos==0)
	{
		pd->drawpixel=!PFPictureElement(pd->gstate, pd->totsiz, pd->ysiz,
													0, pd->ypos);
	}
	if (pd->drawpixel)
	{
		PFColourRGB(pd->gstate,&pd->colour);
		PFPictureElement(pd->gstate, pd->xsiz, pd->ysiz, pd->xpos, pd->ypos);
	}
	pd->xpos+=pd->xsiz;
	pd->row++;
	if (pd->row==pd->width)
	{
		pd->line++;
		pd->row=0;
		pd->xpos=0;
		if (pd->interleaved)
		{
#define INTERLEAVE(x) (int)((height+(x)-1)/(x))
			int height=pd->height;
			int line=pd->line;

			if (line<INTERLEAVE(8))
				pd->ypos=(line*8)*pd->ysiz;
			else if (line<INTERLEAVE(4))
				pd->ypos=((line-INTERLEAVE(8))*8+4)*pd->ysiz;
			else if (line<INTERLEAVE(2))
				pd->ypos=((line-INTERLEAVE(4))*4+2)*pd->ysiz;
			else
				pd->ypos=((line-INTERLEAVE(2))*2+1)*pd->ysiz;
#undef INTERLEAVE
		}
		else
		{
			pd->ypos+=pd->ysiz;
		}
	}
}

/* Process a compression code.	"clear" resets the code table.	Otherwise */
/* make a new code table entry, and output the bytes associated with the */
/* code. */

Error process(picturedata *pd, register int code)
{
	register codetype *p;

	if (code == pd->clear)
	{
		pd->codesize = pd->datasize + 1;
		pd->codemask = (1 << pd->codesize) - 1;
		pd->avail = pd->clear + 2;
		pd->oldcode = -1;
	}
	else if (code < pd->avail)
	{
		outcode(pd,&pd->codetable[code]);
		if (pd->oldcode != -1)
		{
			p = &pd->codetable[pd->avail++];
			p->prefix = &pd->codetable[pd->oldcode];
			p->first = p->prefix->first;
			p->suffix = pd->codetable[code].first;
			if ((pd->avail & pd->codemask) == 0 && pd->avail < 4096)
			{
				pd->codesize++;
				pd->codemask += pd->avail;
			}
		}
		pd->oldcode = code;
	}
	else if (code == pd->avail && pd->oldcode != -1)
	{
		p = &pd->codetable[pd->avail++];
		p->prefix = &pd->codetable[pd->oldcode];
		p->first = p->prefix->first;
		p->suffix = p->first;
		outcode(pd,p);
		if ((pd->avail & pd->codemask) == 0 && pd->avail < 4096)
		{
			pd->codesize++;
			pd->codemask += pd->avail;
		}
		pd->oldcode = code;
	}
	else
	{
		/* illegal code in raster data */
		return ERR_ISYN;
	}
	return ERR_OK;
}


static Error Display(Gstate gstate, char *base, pt xsiz, pt ysiz,
							Error (*ColourSelect)(Gstate, int))
{
	int global; 						/* Is there a global color map? */
	int quit=FALSE;
	unsigned int count, datum=0;
	unsigned char *buf=(unsigned char *)base+sizeof(FileInfo)+6;
	picturedata *pd;
	int bits=0, code;

	/* first check whether some part of the picture is visible ! */
	if (PFPictureElement(gstate,xsiz,ysiz,0,0)) return ERR_OK;

	if (MEMAllocate(sizeof(picturedata),&pd)) return ERR_IMEM;
	pd->gstate=gstate;

	global = buf[4] & 0x80;
	if (global)
	{
		bits = (buf[4] & 0x07) + 1;
		count = 3*(1<<bits);
		MEMCopy(count,buf+7,pd->colourmap);
		buf+=count;
	}
	buf+=7;

	do {
		switch (*buf++) {
		case '\0':
			/* this kludge for non-standard files */
			break;
		case ',':
			pd->width = buf[4] + (buf[5] << 8);
			pd->height = buf[6] + (buf[7] << 8);
			pd->totsiz=xsiz;
			pd->xsiz=xsiz/pd->width;
			pd->ysiz=ysiz/pd->height;
			pd->interleaved = buf[8] & 0x40;

			if (buf[8] & 0x80)
			{
				/* local colour table */
				bits = (buf[8] & 0x7) + 1;
				count = 3*(1<<bits);
				MEMCopy(count,buf+9,pd->colourmap);
				buf+=count;
			}
			buf+=9;

			pd->datasize = *buf++;
			pd->clear = 1 << pd->datasize;
			pd->eoi = pd->clear+1;
			pd->codesize = pd->datasize + 1;
			pd->codemask = (1 << pd->codesize) - 1;
			if (MEMAllocate(4096*sizeof(codetype),&pd->codetable))
			{
				MEMRelease(pd);
				return ERR_IMEM;
			}
			for (code = 0; code < pd->clear; code++)
			{
				pd->codetable[code].prefix = NULL;
				pd->codetable[code].first = code;
				pd->codetable[code].suffix = code;
			}
			bits=0;
			for (count = *buf++; count > 0; count = *buf++)
			{
				for ( ; count-- > 0 ; buf++)
				{
					datum += *buf << bits;
					bits += 8;
					while (bits >= pd->codesize)
					{
						code = datum & pd->codemask;
						datum >>= pd->codesize;
						bits -= pd->codesize;
						if (code == pd->eoi) goto exitloop; /* This kludge */
								 /* because some GIF files aren't standard */
						if (process(pd,code)) goto exitloop;
					}
				}
			}
exitloop:
			MEMRelease(pd->codetable);
			break;
		case ';':
			quit = TRUE;
			break;
		case '!':
			/* Read a GIF extension block (and do nothing with it). */
			buf++;			/* code ?? */
			buf+=*buf+1;	/* length of extension block */
			break;
		default:
			/* illegal GIF block type */
			quit=TRUE;
			break;
		}
	} while (!quit);
	MEMRelease(pd);
	return ERR_OK;
}

PICTdriver init = {
	NULL, PF_PICTUREDRIVER,
	"GIF (87a) Picture driver",
	Recognize,
	AspectRatio,
	Display,
	ColourInfo,
	ColourGet,
	Handle
};
